const os = require('os')
const path = require('path')
const fse = require('fs-extra')
const colors = require('colors')
const semver = require('semver')
const dedent = require('dedent')
const { program } = require('commander')
const packageConfig = require('../package.json')
const {
  LOWEST_NODE_VERSION,
  DEFAULT_CLI_HOME,
  NPM_NAME,
  DEPENDENCIES_PATH
} = require('./const')
const {
  log,
  npm,
  exec, locale, Package
} = require('@jqq-cli-dev/util')

module.exports = cli

let args
let config
const userHome = os.homedir()

async function cli () {
  try {
    await prepare()
    registerCommand()
  } catch (e) {
    // 打印错误调用栈信息
    log.error(e.message)
  }
}

/* 注册命令 */
function registerCommand () {
  program.version(packageConfig.version, '-v, --vers', 'output the current version')
    .usage('<command> [options]')

  // 注册命令 lerna
  program
    .command('lerna')
    .description('学习使用 lerna')
    .action(() => {
      log.success('链接地址', 'lerna使用总结.md')
    })

  // 注册命令 init
  program
    .command('init [type]')
    .description('项目初始化')
    .option('--packagePath <packagePath>', '手动指定 init 包路径')
    .option('--targetPath <packagePath>', '指定模板安装路径')
    .option('--force', '覆盖当前路径文件（谨慎使用）')
    .action(async (type, { packagePath, targetPath, force }) => {
      const packageName = '@jqq-cli-dev/init'
      const packageVersion = '1.0.0' // 写死版本，下载时会下载最新的
      await execCommand({
        packagePath,
        targetPath,
        packageName,
        packageVersion
      }, {
        type,
        force
      })
    })

  // 清除缓存文件
  program
    .command('clean')
    .description('清空缓存文件')
    .option('-a, --all', '清空全部')
    .option('-d, --dep', '清空依赖文件')
    .action((options) => {
      log.notice('开始清空缓存文件')
      if (options.all) {
        // 清空整个安装目录
        cleanAll()
      } else if (options.dep) {
        // 只清除依赖文件
        const depPath = path.resolve(config.cliHome, DEPENDENCIES_PATH)
        if (fse.pathExistsSync(depPath)) {
          fse.emptyDirSync(depPath)
          log.success('清空依赖文件成功', depPath)
        } else {
          log.success('文件夹不存在')
        }
      } else {
        // 没有传 option, 则默认清空安装目录
        cleanAll()
      }
    })

  // 添加 debug 选项
  program
    .option('-d, --debug', '打开调试模式')

  program.parse(process.argv)
}

/* 执行命令 */
async function execCommand ({
  packagePath,
  targetPath = packagePath,
  packageName,
  packageVersion
}, extraOptions) {
  let rootFile
  // log.verbose('execCommand packagePath', packagePath)
  // log.verbose('execCommand targetPath', targetPath)
  // log.verbose('execCommand packageName', packageName)
  // log.verbose('execCommand packageVersion', packageVersion)
  try {
    if (packagePath) {
      // 指定依赖包执行
      const execPackage = new Package({
        targetPath: targetPath,
        storePath: packagePath,
        packageName: packageName,
        packageVersion: packageVersion
      })

      // 获取模板入口文件路径
      rootFile = execPackage.getRootFilePath(true)
    } else {
    // 默认安装路径
      const { cliHome } = config // Users/jqq/.jqq-cli-dev
      const packageDir = `${DEPENDENCIES_PATH}` // dependencies
      const targetPath = path.resolve(cliHome, packageDir) // Users/jqq/.jqq-cli-dev/dependencies
      const storePath = path.resolve(targetPath, 'node_modules') // // Users/jqq/.jqq-cli-dev/dependencies/node_modules
      const initPackage = new Package({
        targetPath,
        storePath,
        packageName: packageName,
        packageVersion: packageVersion
      })

      if (await initPackage.exists()) {
        await initPackage.update()
      } else {
        await initPackage.install()
      }
      rootFile = initPackage.getRootFilePath()
    }
    const _config = Object.assign({}, config, extraOptions, {
      debug: args.debug,
      targetPath
    })

    if (fse.pathExistsSync(rootFile)) {
      // example: 执行 init() 方法
      log.verbose('rootFile', rootFile)
      const code = `require('${rootFile}')(${JSON.stringify(_config)})`

      // node -e 可以直接执行一段js脚本
      const p = exec('node', ['-e', code], {
        stdio: 'inherit'
      })
      p.on('error', (e) => {
        log.verbose('命令执行失败:', e)
        handleError(e)
        process.exit(1)
      })
      p.on('exit', (c) => {
        log.verbose('命令执行结束:', c)
        process.exit(c)
      })
    } else {
      throw new Error('入口文件不存在，请重试')
    }
  } catch (err) {
    log.error(err.message)
  }
}

function handleError (e) {
  if (args.debug) {
    log.error('Error:', e.stack)
  } else {
    log.error('Error:', e.message)
  }
  process.exit(-1)
}

async function prepare () {
  checkPkgVersion()
  checkNodeVersion()
  checkRoot()
  checkUserHome()
  checkInputArgs()
  getEnv()
  await checkGlobalUpdate()
}

/* 打印 cli 版本信息 */
function checkPkgVersion () {
  log.notice('info', '版本号 ' + packageConfig.version)
  log.success(locale.welcome)
}

/* 检查 NodeJs 最低版本 */
function checkNodeVersion () {
  if (!semver.gte(process.version, LOWEST_NODE_VERSION)) {
    throw new Error(colors.inverse(`需要安装 ${LOWEST_NODE_VERSION
      }以上 NODE 版本`))
  }
}

// 如果文件是 root 权限，则当前用户没办法进行修改，所以要进行权限降级
function checkRoot () {
  // sudo jqq-cli-dev
  // console.log(process.env)
  // console.log(process.env.SUDO_UID); //用 sudo 执行命令的当前用户 501
  // console.log(process.getgid()) // group id :0
  // console.log(process.getuid()) // user id:0; jqq-cli-dev => 501; sudo jqq-cli-dev => 0
  const rootCheck = require('root-check')
  rootCheck(colors.red('请避免使用 root 账户启动本应用'))
  // console.log(process.getuid())// user id:501
}

/* 判断用户主目录是否存在 */
function checkUserHome () {
  // console.log(userHome); // /Users/jqq
  // console.log(fse.pathExistsSync(userHome));
  if (!userHome || !fse.pathExistsSync(userHome)) {
    throw new Error(colors.red('当前登录用户的主目录不存在'))
  }
}

/* 格式代命令行参数 */
function checkInputArgs () {
  const minimist = require('minimist')
  args = minimist(process.argv.slice(2))
  setLogLevel(args)
  log.verbose('输入参数', args)
}

/* 检查日志级别 */
function setLogLevel (args) {
  if (args.debug) {
    process.env.LOG_LEVEL = 'verbose'
  } else {
    process.env.LOG_LEVEL = 'info'
  }
  log.level = process.env.LOG_LEVEL
}

/* 获取环境变量 */
function getEnv () {
  log.verbose('开始检查环境变量')

  // 将环境变量中的变量从 .env 文件加载到 process.env 中
  const dotenv = require('dotenv')
  dotenv.config({
    path: path.resolve(userHome, '.env')
  })
  config = createCliConfig()
  log.verbose('环境变量', config)
}

/* 创建 Cli 配置表 */
function createCliConfig (params) {
  const cliConfig = {
    home: userHome
  }
  if (process.env.CLI_HOME) {
    cliConfig.cliHome = path.join(userHome, process.env.CLI_HOME)
  } else {
    cliConfig.cliHome = path.join(userHome, DEFAULT_CLI_HOME)
  }
  return cliConfig
}

async function checkGlobalUpdate () {
  log.verbose('检查 jqq-cli-dev 最新版本')
  const currentVersion = packageConfig.version
  const latestVersion = await npm.getNpmLatestSemverVersion(NPM_NAME, currentVersion)
  if (latestVersion && semver.gt(latestVersion, currentVersion)) {
    log.warn(colors.yellow(dedent(`
    请手动更新 ${NPM_NAME},
    当前版本: ${packageConfig.version}, 最新版本: ${latestVersion}
    更新命令: npm install -g ${NPM_NAME}`)))
  }
}

/* 清除安装目录 */
function cleanAll () {
  if (fse.pathExistsSync(config.cliHome)) {
    fse.emptyDirSync(config.cliHome)
    log.success('清空全部缓存成功', config.cliHome)
  } else {
    log.success(config.cliHome, '文件夹不存在')
  }
}
